#include <msp/io/slice.h>
#include <msp/strings/format.h>
#include <msp/strings/regex.h>
#include "collection.h"
#include "packsource.h"

using namespace std;

namespace {

void delete_io(Msp::IO::Base *io)
{
	delete io;
}

}

namespace Msp {
namespace DataFile {

void PackSource::add_pack_file(const string &fn)
{
	add_pack_file(fn, string());
}

void PackSource::add_pack_file(const string &fn, const string &filter)
{
	add_pack(0, fn, filter);
}

void PackSource::add_pack_io(IO::Seekable &io, const string &fn)
{
	add_pack(&io, fn, string());
}

void PackSource::add_pack_io(IO::Seekable &io, const string &fn, const string &filter)
{
	add_pack(&io, fn, filter);
}

void PackSource::add_pack(IO::Seekable *io, const string &fn, const string &filter)
{
	Pack *pack = 0;
	for(list<Pack>::iterator i=packs.begin(); (!pack && i!=packs.end()); ++i)
		if(i->get_filename()==fn || (io && i->get_io()==io))
			pack = &*i;
	if(!pack)
	{
		packs.push_back(Pack(io, fn));
		pack = &packs.back();
		if(io)
		{
			DataFile::Parser parser(*io, fn);
			Pack::Loader loader(*pack);
			loader.load(parser);
		}
		else
			DataFile::load(*pack, fn);
	}

	FileMap pack_files;
	pack->collect_files(pack_files, filter);
	files.insert(pack_files.begin(), pack_files.end());
	for(FileMap::const_iterator i=pack_files.begin(); i!=pack_files.end(); ++i)
		i->second->collect_objects(objects);
}

list<PackSource::FileInfo> PackSource::list_files() const
{
	list<FileInfo> result;
	for(FileMap::const_iterator i=files.begin(); i!=files.end(); ++i)
		result.push_back(i->second->get_info());
	return result;
}

bool PackSource::is_loadable(const CollectionItemTypeBase &type, const string &name) const
{
	ObjectMap::const_iterator i = objects.find(name);
	if(i==objects.end())
		return false;

	// If the object has a keyword, it must match that of the type
	if(!i->second->get_keyword().empty() && i->second->get_keyword()!=type.get_keyword())
		return false;

	return true;
}

CollectionSource::NameList PackSource::get_names(const CollectionItemTypeBase &type) const
{
	NameList names;
	for(ObjectMap::const_iterator i=objects.begin(); i!=objects.end(); ++i)
	{
		if(!i->second->get_keyword().empty())
		{
			if(i->second->get_keyword()!=type.get_keyword())
				continue;
		}
		else if(!type.match_name(i->first))
			continue;

		names.push_back(i->first);
	}

	return names;
}

void PackSource::load(Collection &coll, const CollectionItemTypeBase &type, const string &name) const
{
	ObjectMap::const_iterator i = objects.find(name);
	if(i==objects.end())
		return;

	File &file = i->second->get_file();
	if(file.is_loaded())
		return;
	file.set_loaded();

	RefPtr<IO::Base> in = file.open();
	Parser parser(*in, file.get_full_name());
	if(file.is_collection())
	{
		Collection::Loader ldr(coll);
		ldr.load(parser);
	}
	else
		type.load_item(coll, parser, name);
}

IO::Seekable *PackSource::open(const string &fn) const
{
	FileMap::const_iterator i = files.find(fn);
	if(i!=files.end())
		return i->second->open().release();

	return 0;
}


PackSource::Pack::Pack(IO::Seekable *i, const string &fn):
	filename(fn),
	io(i),
	base_offset(0)
{ }

void PackSource::Pack::collect_files(FileMap &fm, const string &filter) const
{
	if(filter.empty())
	{
		for(list<File>::const_iterator i=files.begin(); i!=files.end(); ++i)
			fm[i->get_filename()] = &*i;
	}
	else
	{
		Regex re(filter);
		for(list<File>::const_iterator i=files.begin(); i!=files.end(); ++i)
			if(re.match(i->get_filename()))
				fm[i->get_filename()] = &*i;
	}
}


PackSource::File::File(const Pack &p, const string &fn):
	pack(p),
	filename(fn),
	offset(0),
	length(0),
	collection(false),
	loaded(false)
{ }

RefPtr<IO::Seekable> PackSource::File::open() const
{
	if(pack.get_io())
		// TODO Performance may be poor without buffering
		return new IO::Slice(*pack.get_io(), pack.get_base_offset()+offset, length);
	else
	{
		IO::BufferedFile *io_file = new IO::BufferedFile(pack.get_filename());
		IO::Slice *io_slice = new IO::Slice(*io_file, pack.get_base_offset()+offset, length);
		io_slice->signal_deleted.connect(sigc::bind(sigc::ptr_fun(delete_io), io_file));
		return io_slice;
	}
}

PackSource::FileInfo PackSource::File::get_info() const
{
	FileInfo info;
	info.name = filename;
	info.size = length;
	return info;
}

string PackSource::File::get_full_name() const
{
	return format("%s/%s", pack.get_filename(), filename);
}

void PackSource::File::set_loaded()
{
	loaded = true;
}

void PackSource::File::collect_objects(ObjectMap &objs) const
{
	for(list<Object>::const_iterator i=objects.begin(); i!=objects.end(); ++i)
		objs[i->get_name()] = &*i;
}


PackSource::Object::Object(File &f, const string &n, const string &k):
	file(f),
	name(n),
	keyword(k)
{ }


PackSource::Pack::Loader::Loader(Pack &p):
	ObjectLoader<Pack>(p)
{
	add("file",        &Loader::file);
	add("base_offset", &Pack::base_offset);
}

void PackSource::Pack::Loader::file(const string &fn)
{
	obj.files.push_back(File(obj, fn));
	load_sub(obj.files.back());
}


PackSource::File::Loader::Loader(File &f):
	ObjectLoader<File>(f)
{
	add("object", &Loader::object);
	add("slice",  &File::offset, &File::length);
}

void PackSource::File::Loader::finish()
{
	if(!obj.collection)
	{
		PackSource::Object ob(obj, obj.filename, string());
		obj.objects.push_back(ob);
	}
}

void PackSource::File::Loader::object(const string &name, const string &kwd)
{
	obj.objects.push_back(PackSource::Object(obj, name, kwd));
	obj.collection = true;
}

} // namespace DataFile
} // namespace Msp
