#include <msp/strings/format.h>
#include "binaryparser.h"
#include "dataerror.h"
#include "parser.h"
#include "statement.h"
#include "textparser.h"

using namespace std;

namespace Msp {
namespace DataFile {

Parser::Parser(IO::Base &i, const string &s):
	in(i),
	main_src(s),
	src(s),
	good(true),
	mode(new TextParser(in, src))
{ }

Parser::~Parser()
{
	delete mode;
}

Statement Parser::parse(bool raw)
{
	if(!good)
		throw logic_error("Parser::parse() !good");

	try
	{
		while(1)
		{
			Statement st = mode->parse();
			if(!st.keyword.compare(0, 2, "__"))
			{
				st.control = true;
				process_control_statement(st);
			}

			if(raw || !st.control)
				return st;
			else if(!good)  // This will occur with an __end statement
				return Statement();
		}
	}
	catch(const exception &e)
	{
		good = false;
		if(dynamic_cast<const data_error *>(&e))
			throw;
		else
			throw data_error(src, in.get_line_number(), e);
	}
}

void Parser::process_control_statement(const Statement &st)
{
	if(st.keyword=="__bin")
	{
		delete mode;
		mode = new BinaryParser(in, src);

		while(in.peek()=='\n')
			in.get();
	}
	else if(st.keyword=="__text")
	{
		delete mode;
		mode = new TextParser(in, src);
	}
	else if(st.keyword=="__z")
		in.set_decompress();
	else if(st.keyword=="__src")
	{
		string s = st.args[0].get<string>();
		if(s.empty())
			src = main_src;
		else
			src = format("%s[%s]", main_src, s);
	}
	else if(st.keyword=="__end")
		good = false;
	else
		mode->process_control_statement(st);
}

const StatementKey *Parser::peek(unsigned level)
{
	while(good)
	{
		const StatementKey *key = mode->peek(level);
		if(key && !key->keyword.compare(0, 2, "__"))
			process_control_statement(mode->parse());
		else
			return key;
	}

	return 0;
}

bool Parser::parse_and_load(unsigned level, Loader &ldr, const LoaderAction &act)
{
	// Peek first to get any control statements processed
	peek(level);
	return mode->parse_and_load(level, ldr, act);
}

} // namespace DataFile
} // namespace Msp
